% BinProlog x.xx Copyright (C) 1992 Paul Tarau. All rights reserved.
% COMPILER: dcgs --> prolog --> binary progs --> code
% works on a clause at a time, uses no side effects

cutp(X):-name(X,"$cut").

bin_bu(('$bin_cut'(X,Cont):-true(Cont))):-cutp(X).
bin_bu(C):-  
	bu(B,_,Where),
	bu_body(Where,B,C).

bu_body(in_head,B,(B:-true(Cont))):-
	functor(B,_,N),arg(N,B,Cont).
bu_body(in_body,B,(B:-B)).

compile_mem(File):-see(File),compile_mem1(mem,File),seen.

mcompile(File):-
	seeing(X),
	see(File),
	(mcompile1(mem,File)->true;true),
	seen,
	see(X).

mcompile1(Mode,File):-
	mcomp_file(Mode,File),
	terminate_file(mem,'$end1',1).

% Mode must be mem
compile_mem1(Mode,File):-
	ttyprint(compiling(to(Mode),File,'...')),
	ctime(T1),
	mcompile1(Mode,File),
	!,
	ctime(T2),T is T2-T1,
	ttyprint(compile_time(T)),
	abort.
compile_mem1(_,_):-
	ttyprint('compilation aborted'),
	restart,
	abort.

% modes: wam,asm,bin
compile0(Mode,[F|Fs],OutFile):-!,
  ctime(T1),
  xcompile(Mode,[F|Fs],OutFile),
  ctime(T2),
  T is T2-T1,
  write(total_compile_time(T)),nl.
compile0(Mode,File,OutFile):-xcompile(Mode,[File],OutFile).

xcompile(Mode,InFiles,OutFile):-
	tell(OutFile),
	cc_bus(Mode),
	member(InFile,InFiles),
	ttyprint(compiling(to(Mode),InFile,'...')),
	ctime(T1),
	comp_file(Mode,InFile),
	ctime(T2),T is T2-T1,
	ttyprint(compile_time(T)),
	fail.
xcompile(Mode,_,_):-
	terminate_file(Mode,'$end0',0),
	fail.
xcompile(_,_,_):-
	told.

terminate_file(Mode,Dummy,Level):-
	cc(Mode,Dummy),
	emit_code(Mode,[[ii(end,?,Level,Mode)]]).

survive_cleanup(F0,F):-
	name(F0,Survivor),  % Survivor is now a list on the heap
	restart,     % total cleanup of name-spaces, strings, files etc...
	name(F,Survivor),   % F is now a valid name
	see(F),seen.      % F is now linked to an internal file pointer

% compiles to memory	
mcomp_file(Mode,InFile):-
	survive_cleanup(InFile,F),
	translate_all(F,Mode).

comp_file(Mode,InFile):-
	member(Mode,[wam,asm,bin]),!,
	translate_all(InFile,Mode).

translate_all(F,Mode):-
	seeing(F0),
		see(F),
			repeat,
				read(C),
				translate(C,Mode),
			!,
		seen,
	see(F0).

translate(end_of_file,_):-!.
translate(':-'(C),Mode):-!,translate_cmd(C,Mode),fail.
translate('::-'(H,B),Mode):-!,cc_bin(Mode,(H:-B)),fail.
translate(C,Mode):-cc(Mode,C),fail.

translate_cmd([F],Mode):-!,include_file(F,Mode).
translate_cmd(compile(F),Mode):-!,include_file(F,Mode).
translate_cmd(ensure_loaded(F),Mode):-!,include_file(F,Mode).
translate_cmd(consult(F),Mode):-!,include_file(F,Mode).   % can be consult
translate_cmd(reconsult(F),Mode):-!,include_file(F,Mode). % can be consult

translate_cmd(op(X,Y,Z),wam):-!,
	op(X,Y,Z), 
	encode_op(X,Y,Z).
translate_cmd(C,_):-
  telling(F),tell(user),exec_cmd(C),told,tell(F),fail.

exec_cmd(C):-is_compiled(C),!,(C,fail;true).
exec_cmd(C):-errmes(bad_command,C).

include_file(IFile,Mode):-
	seeing(CF1),
	find_file(IFile,F),
	ttyprint(begin(including(F),in(CF1))),
	translate_all(F,Mode),
	seeing(CF2),ttyprint(end(including(F),in(CF2))).

preprocess(Clause,BinClause):-
	preprocess(Clause,_,_,BinClause).

preprocess(C,M,D,B):-
	std_expand_term(C,E),
	fact2rule(E,M),    
	mdef_to_def(M,D),
	def_to_mbin(D,B).

cc(Mode,C):-
	preprocess(C,M,D,B),
	show_steps(Mode,M,D),
	cc_bin(Mode,B).

fact2rule((:-B),(:-B)):-!.
fact2rule((H:-B),(H:-B)):-!.
fact2rule(H,(H:-true)).

% BINARY CLAUSE COMPILER

cc_bus(Mode):-Mode\==bin,bin_bu(B),cc_bin(Mode,B),fail.
cc_bus(_).

cc_bin(bin,C):-!,portray_clause(C),fail.
cc_bin(asm,C):-write('BINARY:'),nl,portray_clause(C),nl,fail.
cc_bin(Mode,C):-
	comp_clause(C,CodeC,Def,Exec),
	emit_code(Mode,[Def,CodeC,Exec]),
	!.
cc_bin(Mode,C):-
	errmes(failing_to_compile_clause(Mode),C).

emit_code(mem,C):-gen_code(mem,C).
emit_code(wam,C):-gen_code(wam,C).
emit_code(asm,C):-show_code(C).

show_steps(asm,M,D):-
  nl,M\==D,write('EXPANDED:'),nl,portray_clause(M),nl,fail
; write('DEFINITE:'),nl,portray_clause(D),nl,fail.
show_steps(_,_,_).
	
comp_clause(C,OCode,
	[ii(clause,?,F1,N1),ii(firstarg,?,G/M,LastN)],
	[ii(execute,?,F2,N2)]):-
	add_true(C,(H:-B)),
	firstarg(H,G/M),
	cc_h_b(H,F1/N1,B,F2/N2,RCode),
	max(N1,N2,MaxN),FirstN is MaxN+1,
	vars(RCode,OCode),
	functor(Dict,dict,MaxN),
	fill_info(OCode,Dict),
	collapse_args(Dict,1,MaxN),
	allocate_regs(OCode,FirstN/FirstN-[],FirstN/LastN-_).

cc_h_b(H,F1/N1,B,F2/N2,RCode):-
	cc_h(H,F1/N1,get-RCode,get-Rest), % pp(h=H),
	cc_b(B,F2/N2,put-Rest,put-[]),!. % pp(b=B)

firstarg(H,FN):-arg(1,H,A),classif_first(A,FN).

classif_first(A,FN):-atomic(A),!,FN=A/0.
classif_first(A,FN):-var(A),!,FN='_'/0.
classif_first(T,F/N):-functor(T,F,N).

cc_h(B,F/N)-->{bu(B,No,in_head)},!,
	{functor(B,F,N),arg(N,B,Cont)},
	emit(get,ii(builtin,?,No,Cont)).
cc_h(T,F/N)-->
	{functor(T,F,N),N>0},!,{functor(CT,F,N)},
	emit_head_top_term(N,T,CT).
cc_h(T,_)-->{errmes(unexpected_head_atom,T)}.

cc_b(Cont,true/1)-->{var(Cont)},!,
	emit_body_top_term(1,true(_),true(Cont)).
cc_b(true,true/0)-->!.
cc_b('$bin_cut'(_cutp,Cont),FN)-->!,
	emit(put,ii(put,_,temp(1),_cutp)),  
	cc_b(Cont,FN).
cc_b(=(A,B,Cont),FN)-->!,
	cc_t(V1=A),
	cc_t(V2=B),
	emit(put,ii(put,_,temp(0),V1)),
	emit(put,ii(get,_,temp(0),V2)), 
	cc_b(Cont,FN).
cc_b(B,FN)-->{bu(B,No,in_body)},!,
	cc_b_bu(No,B,FN).
cc_b(T,F/N)-->
	{functor(T,F,N),N>0},!,{functor(CT,F,N)},
	emit_body_top_term(N,T,CT).
cc_b(T,_)-->{errmes(unexpected_body_atom,T)}.

/*
arith_op(Op):-bu(B,arith(_,_),_),!,functor(B,Op,_).

% arith_no(No):-bu(_,arith(No,_),_).

% arith_outargs(K):-bu(_,arith(_,K),_).
*/

out_reg(0,_,0).
out_reg(1,Res,Res).

cc_b_bu(arith(No,NbOutArgs),OpArgsCont,FN)-->!,
	{ functor(OpArgsCont,Op,N1),arg(N1,OpArgsCont,Cont),
		N is N1-1, arg(N,OpArgsCont,X), out_reg(NbOutArgs,X,Res),
		I is N-NbOutArgs,functor(NewOpArgs,Op,I) % NbOutArgs = 0,1
	},
	handle_constant_res(NbOutArgs,VarRes=Res),
	emit_top_bargs(1,I,OpArgsCont,NewOpArgs),
	emit(put,ii(arith,_Type,No,VarRes)),
	cc_b(Cont,FN).
cc_b_bu(No,BodyAndCont,FN)-->
	{ functor(BodyAndCont,_,N),N1 is N-1,
		arg(N,BodyAndCont,Cont),
		arg(1,BodyAndCont,Arg)
	},
	cc_b_args(N1,Arg),
	emit(put,ii(inline,_,No,_)), % inline=>void variable
	cc_b(Cont,FN).

handle_constant_res(0,_)-->!.
handle_constant_res(1,X=C)-->{var(C)},!,{X=C}.
handle_constant_res(1,X=C)-->{atomic(C)},!,
	emit(put,ii(put,_,temp(0),C)),
	emit(put,ii(get,_,temp(0),X)).
handle_constant_res(1,X=C)-->!,
	cc_t(X=C).

% handle_constant_res(1,_=C)-->{errmes(must_be_atomic_or_var,C)}.

classif_load(X,A,_)-->{var(A)},!,{X=A}.
classif_load(X,A,constant)-->{atomic(A)},!,{X=A}.
classif_load(X,A,_)-->cc_t(X=A).
	
cc_b_args(0,_)-->[].
cc_b_args(1,Arg)-->cc_t(V=Arg),
	emit(put,ii(put,_,temp(0),V)).

emit_top_bargs(I,N,_,_) --> {I>N},!.
emit_top_bargs(I,N,T,CT) --> {I=<N,I1 is I+1},
	{arg(I,T,A),arg(I,CT,X)},
	classif_load(X,A,Type),
	emit(put,ii(load,Type,I,X)),
	emit_top_bargs(I1,N,T,CT).

emit_top_bargs(I,N,T,CT) --> {I=<N},
	{arg(I,T,A),arg(I,CT,X)},
	!,
	cc_t(X=A),
	{I1 is I+1},
	emit_top_bargs(I1,N,T,CT).

emit_head_top_term(N,T,CT) --> 
	emit_top_args(get,1,1,T,CT),
	cc_arg(1,1,CT,T),
	emit_top_args(get,2,N,T,CT),
	cc_arg(2,N,CT,T).
	
emit_body_top_term(N,T,CT) --> 
	cc_arg(1,N,CT,T),!,
	emit_top_args(put,1,N,T,CT).

emit_top_args(_,I,N,_,_) --> {I>N},!.
emit_top_args(Op,I,N,T,CT) --> {I=<N},
  {arg(I,T,A),arg(I,CT,X),classif_arg(X,A,Type)}, % must be int. if const!
  !,
  emit(Op,ii(Op,Type,arg(I),X)),
  {I1 is I+1},
  emit_top_args(Op,I1,N,T,CT).

cc_t(X=T) --> {var(T)},!,{X=T}.
cc_t(X=T) --> {atomic(T)},!,{X=T}.
cc_t(X=T) --> {functor(T,F,N)},{N>0},!,
	{functor(CT,F,N)},
	emit_term(X,F,N,T,CT).

emit_term(X,F,N,T,CT) --> emit_wam(get,X,F,N,T,CT),!,cc_arg(1,N,CT,T).
emit_term(X,F,N,T,CT) --> cc_arg(1,N,CT,T),emit_wam(put,X,F,N,T,CT).

cc_arg(I,N,_,_) -->  {I>N},!.
cc_arg(I,N,CT,T) --> {I=<N},
	{arg(I,T,A),arg(I,CT,X)},
	{I1 is I+1},!,
	cc_t(X=A),
	cc_arg(I1,N,CT,T).

emit_wam(Op,X,F,N,T,CT) --> {N>0},
	emit(Op,ii(Op,structure,F/N,X)),
	emit_args(1,N,T,CT).

emit_args(I,N,_,_) --> {I>N},!.
emit_args(I,N,T,CT) --> {I=<N},
	{arg(I,T,A),arg(I,CT,X),classif_arg(X,A,Type)},
	!,
	emit(Op,ii(UnifyOp,Type,Op,X)),
	{unify_op(Op,UnifyOp),I1 is I+1},
	emit_args(I1,N,T,CT).

unify_op(put,write).
unify_op(get,unify).

classif_arg(X,A,_):-var(A),!,X=A.
classif_arg(X,A,constant):-atomic(A),!,X=A.
classif_arg(_,_,_).

emit(Mode,E,Mode-[E|Es],Mode-Es).

max(X,Y,Z):-X>Y,!,Z=X.
max(_,Y,Y).

add_true((H:-B),(H:-B)):-!.
add_true(H,(H:-true)).


% VARIABLE OCCURRENCE CLASSIFIER

% vars(T,R) :-
% each (selected) variable V of T gives in R a term
%
%   var(NewVar,OccNo/MaxOccurrences)
%
% and T is subject to (an ugly) side effect as selected
% variables get unified to '$OCC'(s(s(...)),MaxOccurrences)

vars(T,R):-
	find_occurrences(T,R,Vars,[]),
	count_occurrences(Vars).

find_occurrences([],[])-->[].
find_occurrences([ii(Op,Type,Val,Var)|L],[ii(Op,Type,Val,Occ)|R])-->
	occurrence(Var,Occ),
	find_occurrences(L,R).

occurrence(A,A)-->{atomic(A)},!.
occurrence(V,var(NewV,1/Max))-->{var(V)},!,
	newvar(X=Max),
	{V='$OCC'(NewV,X=Max)}.
occurrence('$OCC'(OldV,X=Max),var(OldV,K/Max))-->!,
	oldvar(X=Max,K).
occurrence(Var,Occ)-->{errmes(bad_occurrence,at(var=Var,occ=Occ))}.

inc(V,K,K):-var(V),!,V=s(_).
inc(s(V),K1,K3):-K2 is K1+1,inc(V,K2,K3).

oldvar(X=_,K,Xs,Xs):-inc(X,2,K).

newvar(E,[E|Es],Es).

count_occurrences([]):-!.
count_occurrences([X=Max|Vs]):-inc(X,1,Max),count_occurrences(Vs).


% ARGUMENT REGISTER OPTIMIZER

% fills Dict and and marks still free slots in variables
% with information on liftime of arguments

fill_info(Is,Dict):-fill_all(Is,Dict,0,_).

tpoint(T2,T1,T2):-T2 is T1+1. % gets a time-point

fill_all([],_)-->[].
fill_all([I|Is],Dict)-->
	{fill_var_type(I)},
	fill_one(I,Dict),
	fill_all(Is,Dict).

% fills in liftime information using occurrence numbers

fill_var_type(ii(_,Type,_,var(_,Occ))):-var(Type),!,get_var_type(Occ,Type).
fill_var_type(_).

get_var_type(1/_,variable):-!.
get_var_type(K/Max,value):-K=<Max,!.

fill_one(ii(Op,constant,arg(An),_),Dict)-->!,tpoint(T),
	{mark_arg(Op,T,An,var(_-T/T,1/1),Dict)}.
fill_one(ii(_,constant,_,_),_)-->!,tpoint(_).
fill_one(ii(Op,_,arg(An),Xn),Dict)-->!,tpoint(T),
	{mark_arg(Op,T,An,Xn,Dict)},
	{mark_var(T,Xn)}.
fill_one(ii(_,_,_,var(Id,Occ)),_)-->!,tpoint(T),
	{mark_var(T,var(Id,Occ))}.

% marks the argument An of Dict with liftime information
mark_arg(get,From,An,Xn,Dict):-arg(An,Dict,Xn*_-From/_).
mark_arg(put,To  ,An,Xn,Dict):-arg(An,Dict,_*Xn-_/To).

% marks a variable with liftime information
mark_var(T,var(_-T/T,1/1)):-!.
mark_var(T,var(_-T/_,1/Max)):-1<Max,!.
mark_var(T,var(_-_/T,Max/Max)):-1<Max,!.
mark_var(_,var(_-_/_,_/_)).

% collapses arguments and variables, if possible
collapse_args(_,I,Max):-I>Max,!.
collapse_args(Dict,I,Max):-I=<Max,
	arg(I,Dict,X),
	collapse_them(I,X),
	I1 is I+1,
	collapse_args(Dict,I1,Max).

default(V1/V2):-set_to(0,V1),set_to(99999,V2).

set_to(Val,Val):-!.
set_to(_,_).


% checks if argument I living ALife can be collapsed with
%   input head variable H living HLife and 
%   output body variable B living BLife

collapse_them(I,var(H-HLife,_)*var(B-BLife,_)-ALife):-
	default(HLife),default(BLife),default(ALife),
	check_lifetimes(H-HLife,B-BLife,I-ALife).

check_lifetimes(I-HLife,I-BLife,I-ALife):-
	check_var_arg(HLife,ALife),
	check_var_var(HLife,BLife),
	check_var_arg(BLife,ALife),!.
check_lifetimes(I-HLife,_,I-ALife):-
	check_var_arg(HLife,ALife),!.
check_lifetimes(_,I-BLife,I-ALife):-
	check_var_arg(BLife,ALife),!.
check_lifetimes(_,_,_).

check_var_var(_/H2,B1/_):-H2=<B1,!.
check_var_var(H1/_,_/B2):-B2=<H1.

check_var_arg(X1/X2,A1/A2):-
	A1=<X1,X2=<A2.


% TEMPORARY VARIABLE ALOCATOR

allocate_regs([])-->!.
allocate_regs([I|Is])-->
	allocate1(I),
	allocate_regs(Is).

allocate1(ii(_,_,_,var(Reg-_,KMax)))-->allocate_reg(Reg,KMax),!.
allocate1(_)-->[].

allocate_reg(Reg,1/1)-->{var(Reg)},!,
	get_reg(Reg),
	free_reg(Reg).
allocate_reg(Reg,1/Max)-->{var(Reg),1<Max},!,
	get_reg(Reg).
allocate_reg(Reg,Max/Max)-->{1<Max},!,
	free_reg(Reg).

free_reg(Reg,Min/N-Regs,Min/N-[Reg|Regs]):-Reg>=Min,!.
free_reg(_,Rs,Rs).

get_reg(Reg,Min/N-[Reg|Regs],Min/N-Regs):-!.
get_reg(N,Min/N-Regs,Min/N1-Regs):-N1 is N+1.


%----------PREPROCESSOR TO BINARY LOGIC PROGRAMS-------------

% Transforms definite metaprograms to binary meta-programs
% definite meta clause -> definite clause+ some replacements

mdef_to_def((H:-B),(H:-NewB)):-repl_body(B,NewB).
mdef_to_def((:-B),(:-NewB)):-repl_body(B,NewB),!,
	NewB,  % called here and not propagated to the next step
	fail.

% replaces metavars and some builtins in clause bodies
			 
repl_body(MetaVar,call(MetaVar)):-var(MetaVar),!.
repl_body(M,ExpM):-repl_macro(M,ExpM),!.
repl_body(A,NewA):-split_op(A,L),!,strip_nil(L,NewA).
repl_body(A,A).

repl_macro('!','$bin_cut'(X)):-cutp(X).
repl_macro(var(X),Known):-nonvar(X),repl_known(var(X),Known).
repl_macro(nonvar(X),Known):-nonvar(X),repl_known(nonvar(X),Known).
repl_macro(atomic(X),Known):-nonvar(X),repl_known(atomic(X),Known).
repl_macro(float(X),Known):-nonvar(X),repl_known(float(X),Known).
repl_macro(atomic(X),Known):-nonvar(X),repl_known(atomic(X),Known).
repl_macro((A,B),NewAB):-repl_conj(A,B,NewAB).
repl_macro((A;B),NewAB):-repl_disj(A,B,NewAB).
repl_macro((A->B),(NewA->NewB)):-
	repl_body(A,NewA),
	repl_body(B,NewB).
repl_macro(compare(R,A,B),compare0(A,B,R)).
repl_macro(A==B,compare0(A,B,=)).
repl_macro(A@<B,compare0(A,B,<)).
repl_macro(A@>B,compare0(A,B,>)).
repl_macro('=>'(K,X),lval(K1,K2,X)):-repl_lval(K,K1,K2).

repl_conj(A,B,NewAB):-nonvar(A),split_op(A,NewA),!,
	repl_body(B,NewB),
	app_body(NewA,NewB,NewAB).
repl_conj(A,B,(NewA,NewB)):-
	repl_body(A,NewA),
	repl_body(B,NewB).

repl_disj(If,C,if(NewA,NewB,NewC)):-nonvar(If),If=(A->B),!,
	repl_body(A,NewA),
	repl_body(B,NewB),
	repl_body(C,NewC).
repl_disj(A,B,or(NewA,NewB)):-
	repl_body(A,NewA),
	repl_body(B,NewB).

repl_lval('#'(K1,K2),K1,K2):-!.
repl_lval(K,K,K).

repl_known(X,true):-X,!.
repl_known(_,fail).
	
split_op(X is B,R):-split_is_rel(X,B,R).
split_op(A < B,R):-split_rel(less,A,B,R).
split_op(A > B,R):-split_rel(greater,A,B,R).
split_op(A =< B,R):-split_rel(less_eq,A,B,R).
split_op(A >= B,R):-split_rel(greater_eq,A,B,R).
split_op(A =:= B,R):-split_rel(arith_eq,A,B,R).
split_op(A =\= B,R):-split_rel(arith_dif,A,B,R).
% split_op(tab(Expr),R):-split_is(X,Expr,R,[tab(Y)]). % faster static tab

split_is_rel(X,B,[expr(B,X)]):-var(B),!.
split_is_rel(X,B,[+(B,0,X)]):-atomic(B),!.
split_is_rel(X,B,[+(B,0,X)]):-float(B),!.
split_is_rel(X,B,R):-split_is(X,B,R,[]).

app_body([],Bs,Bs):-!.
app_body([A|As],Bs,(A,Cs)):-app_body(As,Bs,Cs).

strip_nil([A],A):-!.
strip_nil([A|As],(A,Bs)):-strip_nil(As,Bs).

split_rel(Op,A,B,Res):-split_rel_1(Op,A,B,Res,[]).

split_rel_1(Op,A,B)-->
	split_is(X,A),
	split_is(Y,B),
	{OpXY=..[Op,X,Y]},
	emit_is(OpXY).

%split_is(X,A)-->{var(A)},!,emit_is(expr(A,X)). % very slow but flexible is/2
split_is(X,A)-->{var(A)},!,{X=A}.               % fast but static is/2
split_is(X,A)-->{atomic(A)},!,{X=A}.
split_is(X,A)-->{float(A)},!,{X=A}.
split_is(R,OpAB)-->
	{OpAB=..[Op,A,B]},!,
	split_is(VA,A),
	split_is(VB,B),
	{OpArgs=..[Op,VA,VB,R]},
	emit_is(OpArgs).
split_is(R,OpA)-->
	{OpA=..[Op,A]},
	split_is(VA,A),
	{OpArgs=..[Op,VA,R]},
	emit_is(OpArgs).

emit_is(X,[X|Xs],Xs).

% converts a definite clause to a binary metaclause
%    where each metavariable Cont represents a "continuation"
%    and a goal G is represented by a clause :- G.

% def_to_mbin((:-B),(:-BC)):-!,add_cont(B,true,BC).
def_to_mbin((H:-B),(HC:-BC)) :- !,
		 add_head_cont(H,Cont,HC), 
		 add_cont(B,Cont,BC).
def_to_mbin(H,(HCont:-true(Cont))):-
		 add_head_cont(H,Cont,HCont).

add_head_cont((G,Gs1),C,GC):-nonvar(Gs1),!,
		 add_head_cont(Gs1,C,Gs2),
		 termcat(G,Gs2,GC).
add_head_cont(G,C,GC):-termcat(G,C,GC).

% adds a continuation to a term

add_cont((true,Gs),C,GC):-!,add_cont(Gs,C,GC).
add_cont((fail,_),C,fail(C)):-!.
add_cont((G,Gs1),C,GC):-!,
		 add_cont(Gs1,C,Gs2),
		 termcat(G,Gs2,GC).
add_cont(G,C,GC):-termcat(G,C,GC).

% ------------------------------------------------------------------
% simple WAM-code lister

show_code(IIs):-
	write('WAM-ASSEMBLER:'),nl,
	member(Is,IIs),
	member(I,Is),
	show_or_skip(I),
	fail.
show_code(_):-nl.

show_or_skip(ii(get,variable,arg(I),var(I-_,_))):-!.
show_or_skip(ii(put,value,arg(I),var(I-_,_))):-!.
show_or_skip(ii(Op,T,X,Y)):-
	write(Op),write('_'),write(T),write(' '),write((X,Y)),nl.
	
% ------------------------------------------------------------------
% BYTE CODE GENERATOR

% eliminates redundancies and beautifies the code
beautify(arg(X),Op,Type,V,To):-!,encode_arg(X,Op,Type,V,To).
beautify(temp(X),Op,Type,V,To):-!,encode_arg(X,Op,Type,V,To).
beautify(Val,Op,Type,var(Xn-_,_),To):-!,encode2(Op,Type,Val,Xn,To).
beautify(put,write,constant,X,To):-cutp(X),!,encode2(push,cut,?,?,To).
beautify(Y,Op,X,Z,To):-encode2(Op,X,Y,Z,To).

encode_arg(I,get,variable,var(I-_,_),_):-!.
encode_arg(I,put,value,var(I-_,_),_):-!.
encode_arg(An,Op,Type,var(Xn-_,_),To):-!,encode2(Op,Type,Xn,An,To).
encode_arg(1,Op,constant,X,To):-cutp(X),!,encode2(Op,cut,?,?,To).
encode_arg(An,Op,constant,S,To):-encode2(Op,constant,S,An,To).

encode2(Op,Type,X,Y,To):-
	symcat(Op,Type,OpType),
	encode(OpType,X,Y,To).

encode(unify_variable,get,Ai,To):-        wcode(To,1,Ai,?,0).
encode(write_variable,put,Ai,To):-        wcode(To,2,Ai,?,0).

encode(unify_value,get,Ai,To):-           wcode(To,3,Ai,?,0).
encode(write_value,put,Ai,To):-           wcode(To,4,Ai,?,0).

encode(unify_constant,get,Const,To):-     wcode(To,5,0,Const,0).
encode(write_constant,put,Const,To):-     wcode(To,6,0,Const,0).

encode(get_constant,Const,Ai,To):-        wcode(To,7,Ai,Const,0).
encode(get_structure,Func/Arity,Ai,To):-  wcode(To,8,Ai,Func,Arity).

encode(put_constant,Const,Ai,To):-        wcode(To,9,Ai,Const,0).
encode(put_structure,Func/Arity,Ai,To):-  wcode(To,10,Ai,Func,Arity).

encode(get_variable,Xn,Ai,To):-           wcode(To,11,Xn,?,Ai). %move_reg
encode(put_value,Xn,Ai,To):-              wcode(To,11,Ai,?,Xn). %move_reg

encode(put_variable,Xn,Ai,To):-           wcode(To,12,Xn,?,Ai).
encode(get_value,Xn,Ai,To):-              wcode(To,13,Xn,?,Ai).

encode(push_cut,?,?,To):-                 wcode(To,14,0,?,0).
encode(put_cut,?,?,To):-                  wcode(To,15,0,?,0).
encode(get_cut,?,?,To):-                  wcode(To,16,0,?,0).

encode('execute_?',Pred,Arity,To):-       wcode(To,17,0,Pred,Arity).
% proceed ==> 18 (by emulator)

encode(load_constant,AReg,Const,To):-     wcode(To,28,AReg,Const,0).

encode(load_variable,AReg,V,To):-         wcode(To,50,AReg,?,V).

encode(load_value,AReg,V,To):-            wcode(To,29,AReg,?,V).

encode('clause_?',Pred,N,To):-n_nop(X1), X is X1+1,          
	wcode(To,X,0,Pred,N).
encode('firstarg_?',F/N,MaxRegs,To):-n_nop(X1), X is X1+2,
	wcode(To,X,MaxRegs,F,N).

encode('end_?',Reg,Mode,To):-             wcode(To,0,Reg,Mode,0).

encode(inline_variable,No,_Arg,To):-n_inline(N),  wcode(To,N,0,?,No).

encode(arith_variable,No,Arg,To):-n_arith(N),   wcode(To,N,Arg,0,No).
encode(arith_value,No,Arg,To):-n_arith(N),      wcode(To,N,Arg,1,No).

encode('builtin_?',No,Arg,To):-n_bu(N),    wcode(To,N,Arg,?,No).

encode_op(P,A,N):-n_nop(X1),  X is X1+3,
	put(X),put(0),put(0),cwrite(P),put(32),cwrite(A),put(32),cwrite(N),put(0).

wcode(mem,Op,Reg,F,N):-add_instr(Op,Reg,F,N).
wcode(wam,Op,Reg,F,N):-put(Op),put(Reg),put(N),
	cwrite(F),
	put(0).

gen_instr(ii(Op,Type,Arg,Var),To,yes):-beautify(Arg,Op,Type,Var,To),!.
gen_instr(_,_,no).

gen_code(To,IIs):-
	member(Is,IIs),member(I,Is),gen_instr(I,To,Ok),
	Ok=no,
	!,
	fail.
gen_code(_,_).

% --------------------------------------------------------------------

make_cmd(Cs,Cmd):-make_cmd1(Cs,List,[]),name(Cmd,List).

make_cmd1([])-->[].
make_cmd1([X|Xs])-->
	{listify(X,L)},
	phrase(L),
	make_cmd1(Xs).

listify([X|Xs],[X|Xs]):-!.
listify(X,L):-name(X,L).

make_appl(File):-
	compile0(wam,[File,'init.pl','lib.pl'],'newappl.bp').
 
% ?- make_executable_unix_appl('/export/home/helios/tarau/b/l38/ru','progs/hello.pl','hello').

make_executable_unix_appl(AbsolutePathToEmulator,SourceF,ExecF):-
	make_appl(SourceF),
	make_cmd([
	"(echo '#! /bin/sh '",
	"; echo 'exec ", 
	AbsolutePathToEmulator, " -q1 $* $0'",
	" ; cat newappl.bp) > ",ExecF,
	" ; chmod +x ", ExecF
	], Cmd),
	system(Cmd).

boot:-make_kernel('wam.bp',[]).
 
make_kernel(File,Extras):-
	kernel_files(Kernel,Extras),
	compile0(wam,Kernel,File).

kernel_files(['oper.pl','init.pl','lib.pl','dcg.pl',
   'read.pl','write.pl','co.pl','top.pl','extra.pl'|Fs],Fs).
	
make:-make(wam).

make(Project):-make(wam,Project).

make(Mode,Project):-
	find_file(Project,F),make_cmd([Project,".bp"],NewF),
	compile0(Mode,F,NewF).

asm:-comp_file(asm,user).

ls:-system('ls -tF').
dir:-system(dir).

t(File):-edit(textedit,File).
e(File):-edit(emacs,File).

edit(Editor,File):-
	find_file(File,F),
        make_cmd([Editor," ",F],R),
        system(R),
	compile_mem(F).

s:-statistics.
