% -------------------------------------------------------------------------
%   File   : INTERP
%   Author : R.A.O'Keefe
%   Updated: 2 March 84
%   Purpose: Meta-circular interpreter for Prolog

/*  This is a genuinely meta-circular interpreter for a subset of Prolog
    containing cuts.  It relies on the fact that disjunction is transparent
    to cut just like conjunction.  If it doesn't work in your Prolog, and
    if you paid more than $100 for it, take your Prolog back to the shop
    and insist that they fix it, there are at least four different ways of
    implementing disjunction so that it works.
*/


/* R.A. O'Keefe's meta-circular interpreter */
/* slightly modified by Paul Tarau */

meta_interpreter(Body):-do_body(Body).

do_body(Body) :-
	do_body(Body, AfterCut, HadCut),
	( HadCut = yes,
		!,
		do_body(AfterCut)
	;   HadCut = no
	).


do_compiled(call(Goal)) :- !,
	nonvar(Goal),
	do_body(Goal).
do_compiled(\+(Goal)) :-
	do_body(Goal),
	!, fail.
do_compiled(\+(_)) :- !.
do_compiled(bagof(X,Y,Z)) :- !,
	bagof(X, do_body(Y), Z).
do_compiled(setof(X,Y,Z)) :- !,
	setof(X, do_body(Y), Z).
do_compiled(findall(X,Y,Z)) :- !,
	findall(X, do_body(Y), Z).
do_compiled(Goal):-Goal.

do_goal(Goal) :-
	is_dynamic(Goal),
	!,
	clause(Goal, Body),	% <--- assume anything else is interpreted
	do_body(Body, AfterCut, HadCut),
	(	HadCut = yes,
		!,
		do_body(AfterCut)
	;	HadCut = no
	).
do_goal(Goal) :-
	is_compiled(Goal), % <--- check for a compiled predicate
	!,
	do_compiled(Goal).
do_goal(Undef):-
	errmes('undefined predicate',Undef).

do_body((Goal,Body), AfterCut, HadCut):-!,
	do_conj(Goal,Body, AfterCut, HadCut).
do_body((Goal;Body), AfterCut, HadCut):-!,
	do_disj(Goal,Body, AfterCut, HadCut).
do_body(!, true, yes).
do_body(Goal, true, no) :- do_goal(Goal).

do_conj(!,AfterCut, AfterCut, yes) :- !.
do_conj(Goal,Body, AfterCut, HadCut) :- 
	do_body(Goal),
	do_body(Body, AfterCut, HadCut).

do_disj(Disj1,_, AfterCut, HadCut) :-
	do_body(Disj1, AfterCut, HadCut).
do_disj(_,Disj2, AfterCut, HadCut) :-
	do_body(Disj2, AfterCut, HadCut).

trace(Goal) :-
	tr_body(Goal, 0).

tr_goal(call(Goal), Depth) :- !,
	nonvar(Goal),
	tr_body(Goal, Depth).
tr_goal(\+(Goal), Depth) :-
	tr_body(Goal, Depth),
	!, fail.
tr_goal(\+(_), _) :- !.
tr_goal(Goal, Depth) :-
	(  tab(Depth), cwrite('Call: '), print(Goal), fail
        ;  is_interactive, cwrite(' <enter=call, other=trace>: '), fail
        ;  true
	),
        (  is_interactive->get0(X),(X=:=10->true;get0(_))
        ;  X=32,nl
        ),
        ( X=:=10->metacall(Goal)
        ; Depth1 is 1+Depth,tr_call(Goal, Depth1)
	),
	( tab(Depth), write('Exit: '), print(Goal), nl, fail
	; true
	; tab(Depth), write('Redo: '), print(Goal), nl, fail
	)
; 	tab(Depth), write('Fail: '), print(Goal), nl, fail.

tr_call(bagof(X,Y,Z), Depth) :- !,	% include these 4 lines if you
	bagof(X, tr_body(Y,Depth), Z).	% really want them, but they do
tr_call(setof(X,Y,Z), Depth) :- !,	% slow things down a bit.
	setof(X, tr_body(Y,Depth), Z).
tr_call(findall(X,Y,Z), Depth) :- !,
	findall(X, tr_body(Y,Depth), Z).
tr_call(Goal, Depth) :-
	is_dynamic(Goal),!,
	clause(Goal, Body),
	tr_body(Body, Depth, AfterCut, HadCut),
	(   HadCut = yes,
		!,
		tab(Depth), write('CUT'), nl,
		tr_body(AfterCut, Depth)
	;   HadCut = no
	).
tr_call(Goal, Depth) :-
	is_compiled(Goal), % <--- checks for compiled predicate
	!, tab(Depth),write(compiled(Goal)),nl,
	Goal.
tr_call(Undef):-
	errmes('undefined predicate',Undef).

tr_body(Body, Depth) :-
	tr_body(Body, Depth, AfterCut, HadCut),
	(   HadCut = yes,
		!,
		tab(Depth), write('CUT'), nl,
		tr_body(AfterCut, Depth)
	;   HadCut = no
	).


tr_body((Conj1,Conj2), Depth, AfterCut, HadCut) :- !,
	tr_body(Conj1, Conj2, Depth, AfterCut, HadCut).
tr_body(!, _, true, yes) :- !.
tr_body((Disj1;_), Depth, AfterCut, HadCut) :-
	tr_body(Disj1, Depth, AfterCut, HadCut).
tr_body((_;Disj2), Depth, AfterCut, HadCut) :- !,
	tr_body(Disj2, Depth, AfterCut, HadCut).
tr_body(true, _, true, no) :- !.
tr_body(Goal, Depth, true, no) :-
	tr_goal(Goal, Depth).

tr_body(!, AfterCut, _, AfterCut, yes) :- !.
tr_body((A,B), Conj, Depth, AfterCut, HadCut) :- !,
	tr_body(A, (B,Conj), Depth, AfterCut, HadCut).
tr_body((Disj1;_), Conj, Depth, AfterCut, HadCut) :-
	tr_body(Disj1, Conj, Depth, AfterCut, HadCut).
tr_body((_;Disj2), Conj, Depth, AfterCut, HadCut) :- !,
	tr_body(Disj2, Conj, Depth, AfterCut, HadCut).
tr_body(true, Body, Depth, AfterCut, HadCut) :- !,
	tr_body(Body, Depth, AfterCut, HadCut).
tr_body(Goal, Body, Depth, AfterCut, HadCut) :-
	tr_goal(Goal, Depth),
	tr_body(Body, Depth, AfterCut, HadCut).



/* dynamic/1, assert/1 & retract/1 predicates

This is an approximation of other Prologs assert & retract predicates.
For efficiency and programming style reasons we strongly suggest
not to use them too much.

If you want maximal efficiency use bb_def/3 bb_set/3 bb_val/3 
and lval/3. They give you acces to a very fast
hashing table <key,key>--> value, the same that BinProlog
uses internally for indexing by predicate and first argument.

Lval/3 (i,i,?) introduces named global logical variables.
A friendlier syntax (with no extra cost) is

Key1#Key2=>Var  ...for lval(Key1,Key2,Var)

and Key=>Var    ...for lval(Key,Key,Var)

Beware that assert & retract are not optimized for large databases
or frequent use of an asserted predicate.

To use dynamic predicates you have to declare them with dynamic/1
otherwise asserts will simply fail.

To activate an asserted predicate you must use

	?-metacall(Goal).

instead of 

	?- Goal.

To call a dynamic predicate from compiled code, use also metacall/1. 
*/

dynamic(Ps):-make_dynamic(Ps),fail.
dynamic(_).
 
make_dynamic((P1,P2)):-!,make_dynamic(P1),make_dynamic(P2).
make_dynamic(P/N):-
	functor(T,P,N),
	check_dynamic(T,P/N),
	define_dynamic(T).

check_dynamic(T,PN):-is_dynamic(T),!,errmes('already dynamic',PN),fail.
check_dynamic(T,PN):-is_compiled(T),!,
	write('WARNING: redefining compiled predicate'(PN)),nl.
check_dynamic(_,_).

define_dynamic(Pred):-bb_def(Pred,'$first',0),bb_def(Pred,'$last',0).

add_clause(AZ,Inc,H,B):-
	val(H,AZ,V),
	val(H,V,_),
	!,
	V1 is V+Inc,
	bb_def(H,V1,(H:-B)),
	set(H,AZ,V1).
add_clause(AZ,Inc,H,B):-
	val(H,AZ,V),
	bb_def(H,V,(H:-B)).

asserta(C):-
	add_true(C,(H:-B)),
	add_clause('$first',-1,H,B).
	
assertz(C):-
	add_true(C,(H:-B)),
	add_clause('$last',1,H,B).

assert(C):-assertz(C).

retract(C0):-
	add_true(C0,C),C=(H:-B),
	clause0(H,B,I),
	erase0(H,I).

retractall(C0):-
	add_true(C0,(H:-_)),
	val(H,'$first',Min),
	val(H,'$last',Max),
	for(I,Min,Max),
	erase0(H,I),
	fail.
retractall(_).

collect_slot(H,Min):-
	val(H,'$first',Min),
	val(H,'$last',Max),
	Min<Max,!,
	Min1 is Min+1,
	set(H,'$first',Min1).
collect_slot(_,_).

instance('$ref'(H0,I),C):-val(H0,I,C0),copy_term(C0,C).

erase('$ref'(H,I)):-erase0(H,I).

erase0(H,I):-bb_rm(H,I),collect_slot(H,I).

clause(H,B):-clause0(H,B,_).

clause(H,B,'$ref'(H,I)):-clause0(H,B,I).

clause0(H,B,I):-
	val(H,'$first',Min),
	val(H,'$last',Max),
	findall(IC,ith_clause(H,IC,Min,Max),ICs),
	member(I-(H:-B),ICs).

ith_clause(G,Min-C,Min,Max):- % Min=<Max,
	val(G,Min,C).
ith_clause(G,IC,Min,Max):-
        Min<Max,
        Min1 is Min+1,
        ith_clause(G,IC,Min1,Max).

abolish(F,N):-
	functor(P,F,N),functor(C,F,N),
	retractall(C),
	fail.
abolish(F,N):-
	functor(P,F,N),
	set(P,'$first',0),
	set(P,'$last',0).

ensure_dynamic(H,_,_):-is_dynamic(H),!.
ensure_dynamic(_,P,N):-make_dynamic(P/N).

assert_it(end_of_file):-!.
assert_it(C):-
	add_true(C,(H:-_)),
	functor(H,P,N),
	ensure_dynamic(H,P,N),
	assertz(C),!.
assert_it(C):-errmes('unable to assert',C).


% consulting `interpreted code'

debug(F):-reconsult(F).

consult_warning:-write('% using compile/1 is MUCH faster'),nl.

consult(File):-
	find_file(File,F),
	consult_warning,
        consult0(F).

reconsult(File):-
	find_file(File,F0),
	consult_warning,
	survive_cleanup(F0,F),
        consult0(F).

consult0(F):-
	ctime(T1),
	write(consulting(F)),nl,
	see(F),
	repeat,
		read(C),consult_embedded(C),C=end_of_file,!,
	seen,
	ctime(T2),T is T2-T1,
	write(consulted(F,time=T)),nl.

consult_included(F):-seeing(G),find_file(F,F0),consult0(F0),see(G).

consult_embedded(':-'(Cmd)):-!,consult_cmd(Cmd).
consult_embedded(C):-expand_term(C,E),assert_it(E).

consult_cmd([F]):-!,consult_included(F).
consult_cmd(consult(F)):-!,consult_included(F).
consult_cmd(reconsult(F)):-!,consult_included(F).
consult_cmd(C):-metacall(C).

listing(P,N):-
	functor(H,P,N),
	clause(H,B),
        portray_clause((H:-B)),
        fail
;	nl.

listing:-
	bb_list(0,Xs),
	bb_element(P/N+'$first'/0=_,Xs),
	listing(P,N),
	fail
;       nl.

% Blackboard related utilities

% bboard visualisation

bb_element(P/N+F/K=V,[P,N,F,K,V|_]).
bb_element(D,[_,_,_,_,_|L]):-bb_element(D,L).

bb_list(L):-bb_list(0,L). % lists for arity >= 0

bb:-
	statistics(bboard,X),write(bboard-X),nl,
	bb_list(L),
		bb_element(B,L),
		write(B),nl,
	fail
; nl.

/* new stack package, working fine with bb_gc */

bb_def_stack(Stack):-bb_def(Stack,'$top',0).

bb_push(Stack,Ob):-
	val(Stack,'$top',I),!,
	I1 is I+1,
	bb_def(Stack,I1,Ob),
	bb_set(Stack,'$top',I1).
bb_push(Stack,_):-
	errmes('undefined stack',Stack).

bb_pop(Stack,Ob):-
	bb_val(Stack,'$top',I),
	bb_val(Stack,I,Ob),
	I>0,I1 is I-1,!,
	bb_rm(Stack,I),
	bb_set(Stack,'$top',I1).
bb_pop(Stack,_):-
	errmes('stack underflow',stack(Stack)).

bb_stack(Stack,List):-
	val(Stack,'$top',Top),
	findall(X,bb_stack_member(Stack,Top,X),List).

bb_stack_member(Stack,Max,Elem):-
	for(I,1,Max),
	val(Stack,I,Elem).

bb_stack_rm(Stack):-
	val(Stack,'$top',Top),bb_rm(Stack,'$top'),
	for(I,1,Top),
	bb_rm(Stack,I),
	fail.
bb_stack_rm(_).


% unsupported & unsafe

let(X,Y,V):-def(X,Y,V),!.
let(X,Y,V):-set(X,Y,V).

def(A,X):-def(A,A,X).

set(A,X):-set(A,A,X).

val(A,X):-val(A,A,X).

let(A,X):-let(A,A,X).

rm(A):-rm(A,A).

/* old stacks: for comatibility only */

push(Type,S,X):-val(Type,S,Xs),!,bb_set(Type,S,[X|Xs]).
push(Type,S,X):-bb_def(Type,S,[X]).

pop(Type,S,X):-val(Type,S,[A|Xs]),set(Type,S,Xs),copy_term(A,X).

stack(Type,S,Xs):-val(Type,S,Xs).

push(S,X):-push('$stack',S,X).

pop(S,X):-pop('$stack',S,X).

stack(S,X):-stack('$stack',S,X).


/*******************************************************************/

% sort, adapted from public domain code written by R.A. O'Keefe
% use merge_sort(<,_,_) if you do not want duplications eliminated
% use merge_sort(>,_,_) for descending order

sort(L1,L2):-merge_sort(<,L1,DupL),remdup(DupL,L2).

remdup([],[]).
remdup([X,Y|Xs],Ys):-compare(=,X,Y),!,remdup([X|Xs],Ys).
remdup([X|Xs],[X|Ys]):-remdup(Xs,Ys).
      
merge_sort(Rel, L,S ):-
	length(L,N),
	merge_sort1(N, Rel, L,S,[] ).

merge_sort1( 0,_,L,[],L ):-!.
merge_sort1( 1,_,[X|L],[X],L ):-!.
merge_sort1( N,Rel,L,S,R ):-	% N >= 2
	N1 is N >> 1,	N2 is N-N1,
	merge_sort1( N1,Rel,L,S1,R1),	
	merge_sort1( N2,Rel,R1,S2,R),
	merge_2( S2,Rel,S1,S ).

merge_2([],_,S,S ):-!.
merge_2([X|L1],Rel,[Y|L2],[X|L] ):-compare(Rel,X,Y),!,
	merge_2(L1,Rel,[Y|L2],L ).
merge_2(L1,Rel,[Y|L2],[Y|L] ):-
	merge_2(L2,Rel,L1,L ).

%   Keysorting.  Adapted by Mats Carlsson from R.O'Keefe's code, 
%		but uses recursion instead of an auxiliary stack.  
%		Takes care to check validity of arguments.
%   Could be speed up if there were an inline keycompare/3.

ksort(List, Sorted) :-
	keysort(List, -1, S, []), !,
	Sorted = S.
ksort(X, Y):-user_error('illegal_arguments',keysort(X,Y)).

keysort([Head|Tail], Lim, Sorted, Rest) :- !,
	nonvar(Head),
	Head = _-_,
	Qh = [Head|_],
	samkeyrun(Tail, Qh, Qh, Run, Rest0),
	keysort(Rest0, 1, Lim, Run, Sorted, Rest).
keysort(Rest, _, [], Rest).

keysort([Head|Tail], J, Lim, Run0, Sorted, Rest) :-
	J =\= Lim, !,
	nonvar(Head),
	Head = _-_,
	Qh = [Head|_],
	samkeyrun(Tail, Qh, Qh, Run1, Rest0),
	keysort(Rest0, 1, J, Run1, Run2, Rest1),
	keymerge(Run0, Run2, Run),
	K is J+J,
	keysort(Rest1, K, Lim, Run, Sorted, Rest).
keysort(Rest, _, _, Sorted, Sorted, Rest).

samkeyrun([Hd|Tail], QH, QT, Run, Rest) :-
	nonvar(Hd),
	Hd = H-_,
	QT = [Q-_|QT2], 
	Q @=< H, !,
	QT2 = [Hd|_],
	samkeyrun(Tail, QH, QT2, Run, Rest).
samkeyrun([Hd|Tail], QH, QT, Run, Rest) :-
	nonvar(Hd),
	Hd = H-_,
	QH = [Q-_|_],
	H @< Q, !,
	samkeyrun(Tail, [Hd|QH], QT, Run, Rest).
samkeyrun(Rest, Run, [_], Run, Rest).

% keymerge(+List, +List, -List).
keymerge([], L2, Out) :- !,
	Out = L2.
keymerge([H1|T1], L2, Out) :-	
	L2 = [K2-_|_],
	H1 = K1-_,
	K1 @=< K2, !,
	Out = [H1|Out1],
	keymerge(T1, L2, Out1).
keymerge(L1, [H2|L2], Out) :- !,
	Out = [H2|Out1],
	keymerge(L1, L2, Out1).
keymerge(List, _, List).

% -------------------- ---------------------

%   File   : SETOF.PL
%   Author : R.A.O'Keefe
%   Updated: 17 November 1983
%   Purpose: define setof/3, bagof/3, findall/3, and findall/4
%   Needs  : Not.Pl

% Adapted by Paul Tarau for BinProlog: uses a heap based findall/3.
% therefore some database hacking predicates have been removed.
% Updated: 19 July 1992

/*  This file defines two predicates which act like setof/3 and bagof/3.
    I have seen the code for these routines in Dec-10 and in C-Prolog,
    but I no longer recall it, and this code was independently derived
    in 1982 by me and me alone.

    Most of the complication comes from trying to cope with free variables
    in the Filter; these definitions actually enumerate all the solutions,
    then group together those with the same bindings for the free variables.
    There must be a better way of doing this.  I do not claim any virtue for
    this code other than the virtue of working.  In fact there is a subtle
    bug: if setof/bagof occurs as a data structure in the Generator it will
    be mistaken for a call, and free variables treated wrongly.  Given the
    current nature of Prolog, there is no way of telling a call from a data
    structure, and since nested calls are FAR more likely than use as a
    data structure, we just put up with the latter being wrong.  The same
    applies to negation.

    Would anyone incorporating this in their Prolog system please credit
    both me and David Warren;  he thought up the definitions, and my
    implementation may owe more to subconscious memory of his than I like
    to think.  At least this ought to put a stop to fraudulent claims to
    having bagof, by replacing them with genuine claims.

    Thanks to Dave Bowen for pointing out an amazingly obscure bug: if
    the Template was a variable and the Generator never bound it at all
    you got a very strange answer!  Now fixed, at a price.

	bagof/3,		%   Like bagof (Dec-10 manual p52)
	setof/3.		%   Like setof (Dec-10 manual p51)

:- mode
	bagof(+,+,?),
	concordant_subset(+,+,-),
	concordant_subset(+,+,-,-),
	concordant_subset(+,+,+,+,-),
	replace_key_variables(+,+,+),
	setof(+,+,?).

%   setof(Template, Generator, Set)
%   finds the Set of instances of the Template satisfying the Generator..
%   The set is in ascending order (see compare/3 for a definition of
%   this order) without duplicates, and is non-empty.  If there are
%   no solutions, setof fails.  setof may succeed more than one way,
%   binding free variables in the Generator to different values.  This
%   predicate is defined on p51 of the Dec-10 Prolog manual.
*/

setof(Template, Filter, Set) :-
	bagof(Template, Filter, Bag),
	sort(Bag, Set).


%   bagof(Template, Generator, Bag)
%   finds all the instances of the Template produced by the Generator,
%   and returns them in the Bag in they order in which they were found.
%   If the Generator contains free variables which are not bound in the
%   Template, it assumes that this is like any other Prolog question
%   and that you want bindings for those variables.  (You can tell it
%   not to bother by using existential quantifiers.)
%   bagof records three things under the key '.':
%	the end-of-bag marker	       -
%	terms with no free variables   -Term
%	terms with free variables   Key-Term
%   The key '.' was chosen on the grounds that most people are unlikely
%   to realise that you can use it at all, another good key might be ''.
%   The original data base is restored after this call, so that setof
%   and bagof can be nested.  If the Generator smashes the data base
%   you are asking for trouble and will probably get it.
%   The second clause is basically just findall, which of course works in
%   the common case when there are no free variables.

bagof(Template, Generator, Bag) :-
	free_variables(Generator, Template, [], Vars),
	Vars \== [],
	!,
	Key =.. [.|Vars],
	functor(Key, ., N),
	findall(Key-Template,Generator,Recorded),
	replace_instance(Recorded, Key, N, [], OmniumGatherum),
	keysort(OmniumGatherum, Gamut), !,
	concordant_subset(Gamut, Key, Answer),
	Bag = Answer.
bagof(Template, Generator, [B|Bag]) :-
	findall(Template,Generator,[B|Bag]).

_^Goal:-Goal.

replace_instance([], _, _, AnsBag, AnsBag) :- !.
replace_instance([NewKey-Term|Xs], Key, NVars, OldBag, NewBag) :-
		replace_key_variables(NVars, Key, NewKey), !,
		replace_instance(Xs,Key, NVars, [NewKey-Term|OldBag], NewBag).

%   There is a bug in the compiled version of arg in Dec-10 Prolog,
%   hence the rather strange code.  Only two calls on arg are needed
%   in Dec-10 interpreted Prolog or C-Prolog.

replace_key_variables(0, _, _) :- !.
replace_key_variables(N, OldKey, NewKey) :-
	arg(N, NewKey, Arg),
	nonvar(Arg), !,
	M is N-1,
	replace_key_variables(M, OldKey, NewKey).
replace_key_variables(N, OldKey, NewKey) :-
	arg(N, OldKey, OldVar),
	arg(N, NewKey, OldVar),
	M is N-1,
	replace_key_variables(M, OldKey, NewKey).

%   concordant_subset([Key-Val list], Key, [Val list]).
%   takes a list of Key-Val pairs which has been keysorted to bring
%   all the identical keys together, and enumerates each different
%   Key and the corresponding lists of values.

concordant_subset([Key-Val|Rest], Clavis, Answer) :-
	concordant_subset(Rest, Key, List, More),
	concordant_subset(More, Key, [Val|List], Clavis, Answer).


%   concordant_subset(Rest, Key, List, More)
%   strips off all the Key-Val pairs from the from of Rest,
%   putting the Val elements into List, and returning the
%   left-over pairs, if any, as More.

concordant_subset([Key-Val|Rest], Clavis, [Val|List], More) :-
	Key == Clavis,
	!,
	concordant_subset(Rest, Clavis, List, More).
concordant_subset(More, _, [], More).


%   concordant_subset/5 tries the current subset, and if that
%   doesn't work if backs up and tries the next subset.  The
%   first clause is there to save a choice point when this is
%   the last possible subset.

concordant_subset([],   Key, Subset, Key, Subset) :- !.
concordant_subset(_,    Key, Subset, Key, Subset).
concordant_subset(More, _,   _,   Clavis, Answer) :-
	concordant_subset(More, Clavis, Answer).


% ---extracted from: not.pl --------------------%

%   Author : R.A.O'Keefe
%   Updated: 17 November 1983
%   Purpose: "suspicious" negation 

%   In order to handle variables properly, we have to find all the 
%   universally quantified variables in the Generator.  All variables
%   as yet unbound are universally quantified, unless
%	a)  they occur in the template
%	b)  they are bound by X^P, setof, or bagof
%   free_variables(Generator, Template, OldList, NewList)
%   finds this set, using OldList as an accumulator.

free_variables(Term, Bound, VarList, [Term|VarList]) :-
	var(Term),
	term_is_free_of(Bound, Term),
	list_is_free_of(VarList, Term),
	!.
free_variables(Term, _, VarList, VarList) :-
	var(Term),
	!.
free_variables(Term, Bound, OldList, NewList) :-
	explicit_binding(Term, Bound, NewTerm, NewBound),
	!,
	free_variables(NewTerm, NewBound, OldList, NewList).
free_variables(Term, Bound, OldList, NewList) :-
	functor(Term, _, N),
	free_variables(N, Term, Bound, OldList, NewList).

free_variables(0,    _,     _, VarList, VarList) :- !.
free_variables(N, Term, Bound, OldList, NewList) :-
	arg(N, Term, Argument),
	free_variables(Argument, Bound, OldList, MidList),
	M is N-1, !,
	free_variables(M, Term, Bound, MidList, NewList).

%   explicit_binding checks for goals known to existentially quantify
%   one or more variables.  In particular \+ is quite common.

explicit_binding(\+ _,	       Bound, fail,	Bound      ).
explicit_binding(not(_),	     Bound, fail,	Bound	   ).
explicit_binding(Var^Goal,	   Bound, Goal,	Bound+Var).
explicit_binding(setof(Var,Goal,Set),  Bound, Goal-Set, Bound+Var).
explicit_binding(bagof(Var,Goal,Bag),  Bound, Goal-Bag, Bound+Var).

term_is_free_of(Term, Var) :-
	var(Term), !,
	Term \== Var.
term_is_free_of(Term, Var) :-
	functor(Term, _, N),
	term_is_free_of(N, Term, Var).

term_is_free_of(0, _, _) :- !.
term_is_free_of(N, Term, Var) :-
	arg(N, Term, Argument),
	term_is_free_of(Argument, Var),
	M is N-1, !,
	term_is_free_of(M, Term, Var).

list_is_free_of([], _).
list_is_free_of([Head|Tail], Var) :-
	Head \== Var,
	list_is_free_of(Tail, Var).

keysort(L,S):-ksort(L,S).


% for compatibility with various prologs

% SOUND negation -> replaced with (I think) an improved one

% REPLACED:
% not(X):-ground(X),!, \+ X.
% not(X):-user_error('should be ground',not(X)).

% I see no reason to prohibit free variables when the negation succeeds
% as no future bindings can change the logical meaning in this case.
% the case of failure is (of course) different

not(X):- \+ X,!.
not(X):-ground(X),!,fail.
not(X):-user_error('should be ground',not(X))->fail;abort.

term_chars(T,Cs):-nonvar(T),!,swrite(T,S),name(S,Cs).
term_chars(T,Cs):-name(S,Cs),sread(S,T).

gensym(Root,Symbol):-
        val(gensym,Root,N),!,
        N1 is N+1,
        symcat(Root,N1,Symbol),
        set(gensym,Root,N1).
gensym(Root,Symbol):-
        def(gensym,Root,1),
        symcat(Root,1,Symbol).
